/*
// @HEADER
// ***********************************************************************
//
//          Tpetra: Templated Linear Algebra Services Package
//                 Copyright (2008) Sandia Corporation
//
// Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
// the U.S. Government retains certain rights in this software.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
// 1. Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
//
// 2. Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
//
// 3. Neither the name of the Corporation nor the names of the
// contributors may be used to endorse or promote products derived from
// this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY SANDIA CORPORATION "AS IS" AND ANY
// EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL SANDIA CORPORATION OR THE
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// Questions? Contact Michael A. Heroux (maherou@sandia.gov)
//
// ************************************************************************
// @HEADER
*/

#include "Tpetra_Map.hpp"
#include "Tpetra_TestingUtilities.hpp"

// Bug 6437 test
//
// Exercise Tpetra::Map::getLocalMap().  Make sure that the returned
// "local Map" object behaves like Tpetra::Map, for the methods that
// the two have in common.

namespace { // (anonymous)

using Tpetra::TestingUtilities::getDefaultComm;
using Teuchos::Comm;
using Teuchos::outArg;
using Teuchos::RCP;
using Teuchos::REDUCE_MIN;
using Teuchos::reduceAll;
using std::endl;
typedef Tpetra::global_size_t GST;

//
// Test LocalMap with a uniform contiguous Map.
//
TEUCHOS_UNIT_TEST_TEMPLATE_3_DECL( LocalMap, UniformContig, LO, GO, NT )
{
  typedef Tpetra::Map<LO, GO, NT> map_type;
  int lclSuccess = 1;
  int gblSuccess = 1;

  Teuchos::OSTab tab0 (out);
  out << "LocalMap: Uniform Contiguous Map" << endl;
  Teuchos::OSTab tab1 (out);

  RCP<const Comm<int> > comm = getDefaultComm ();
  const LO numLclElts = 10;

  const GST numGblElts = static_cast<GST> (comm->getSize () * numLclElts);
  const GO indexBase = 0;

  map_type gblMap (numGblElts, indexBase, comm);
  auto lclMap = gblMap.getLocalMap ();

  try {
    // Sanity check on the global Map.
    TEST_EQUALITY( gblMap.getNodeNumElements (), static_cast<size_t> (numLclElts) );
    TEST_EQUALITY( gblMap.getMinLocalIndex (), static_cast<LO> (0) );
    TEST_EQUALITY( gblMap.getMaxLocalIndex (), static_cast<LO> (numLclElts - 1) );

    // Test constants.
    TEST_EQUALITY( gblMap.getNodeNumElements (), static_cast<size_t> (lclMap.getNodeNumElements ()) );
    TEST_EQUALITY( gblMap.getIndexBase (), lclMap.getIndexBase () );
    TEST_EQUALITY( gblMap.isContiguous (), lclMap.isContiguous () );
    TEST_EQUALITY( gblMap.getMinLocalIndex (), lclMap.getMinLocalIndex () );
    TEST_EQUALITY( gblMap.getMaxLocalIndex (), lclMap.getMaxLocalIndex () );
    TEST_EQUALITY( gblMap.getMinGlobalIndex (), lclMap.getMinGlobalIndex () );
    TEST_EQUALITY( gblMap.getMaxGlobalIndex (), lclMap.getMaxGlobalIndex () );

    // Test global-to-local and local-to-global index conversion.
    if (numLclElts > 0) {
      for (LO lid = 0; lid < numLclElts; ++lid) {
        const GO expectedGid = gblMap.getGlobalElement (lid);
        // Make sure that the (global) Map behaves as expected.
        TEST_INEQUALITY( expectedGid, Tpetra::Details::OrdinalTraits<GO>::invalid () );
        TEST_EQUALITY( gblMap.getLocalElement (expectedGid), lid );
        // Test the local Map.
        TEST_EQUALITY( lclMap.getGlobalElement (lid), expectedGid );
        TEST_EQUALITY( lclMap.getLocalElement (expectedGid), lid );
      }
    }
  } catch (...) {
    // None of the above methods should throw exceptions, but we catch
    // just in case.
    success = false;
  }

  // Make sure that all processes succeeded.
  lclSuccess = success ? 1 : 0;
  reduceAll<int, int> (*comm, REDUCE_MIN, lclSuccess, outArg (gblSuccess));
  TEST_EQUALITY_CONST( gblSuccess, 1 );

  if (gblSuccess != 1) {
    out << "The comparison test between Map and LocalMap "
      "failed on at least one process." << endl;
    // Make sure that the test fails, even if Process 0 was fine.
    success = false;
    return;
  }
}

//
// Test LocalMap with a NONuniform contiguous Map.
//
TEUCHOS_UNIT_TEST_TEMPLATE_3_DECL( LocalMap, NonuniformContig, LO, GO, NT )
{
  typedef Tpetra::Map<LO, GO, NT> map_type;
  int lclSuccess = 1;
  int gblSuccess = 1;

  Teuchos::OSTab tab0 (out);
  out << "LocalMap: Nonuniform Contiguous Map" << endl;
  Teuchos::OSTab tab1 (out);

  RCP<const Comm<int> > comm = getDefaultComm ();
  const int myRank = comm->getRank ();

  // Process p gets 5+p indices.
  const LO numLclElts = static_cast<LO> (5 + myRank);
  const GST numGblElts = Tpetra::Details::OrdinalTraits<GST>::invalid ();
  const GO indexBase = 0;

  map_type gblMap (numGblElts, static_cast<size_t> (numLclElts), indexBase, comm);
  auto lclMap = gblMap.getLocalMap ();

  try {
    // Sanity check on the global Map.
    TEST_EQUALITY( gblMap.getNodeNumElements (), static_cast<size_t> (numLclElts) );
    TEST_EQUALITY( gblMap.getMinLocalIndex (), static_cast<LO> (0) );
    TEST_EQUALITY( gblMap.getMaxLocalIndex (), static_cast<LO> (numLclElts - 1) );

    // Test constants.
    TEST_EQUALITY( gblMap.getNodeNumElements (), static_cast<size_t> (lclMap.getNodeNumElements ()) );
    TEST_EQUALITY( gblMap.getIndexBase (), lclMap.getIndexBase () );
    TEST_EQUALITY( gblMap.isContiguous (), lclMap.isContiguous () );
    TEST_EQUALITY( gblMap.getMinLocalIndex (), lclMap.getMinLocalIndex () );
    TEST_EQUALITY( gblMap.getMaxLocalIndex (), lclMap.getMaxLocalIndex () );
    TEST_EQUALITY( gblMap.getMinGlobalIndex (), lclMap.getMinGlobalIndex () );
    TEST_EQUALITY( gblMap.getMaxGlobalIndex (), lclMap.getMaxGlobalIndex () );

    // Test global-to-local and local-to-global index conversion.
    if (numLclElts > 0) {
      for (LO lid = 0; lid < numLclElts; ++lid) {
        const GO expectedGid = gblMap.getGlobalElement (lid);
        TEST_EQUALITY( lclMap.getGlobalElement (lid), expectedGid );
        TEST_EQUALITY( lclMap.getLocalElement (expectedGid), lid );
      }
    }
  } catch (...) {
    // None of the above methods should throw exceptions, but we catch
    // just in case.
    success = false;
  }

  // Make sure that all processes succeeded.
  lclSuccess = success ? 1 : 0;
  reduceAll<int, int> (*comm, REDUCE_MIN, lclSuccess, outArg (gblSuccess));
  TEST_EQUALITY_CONST( gblSuccess, 1 );

  if (gblSuccess != 1) {
    out << "The comparison test between Map and LocalMap "
      "failed on at least one process." << endl;
    // Make sure that the test fails, even if Process 0 was fine.
    success = false;
    return;
  }
}

//
// Test LocalMap with a NONcontiguous Map.
//
TEUCHOS_UNIT_TEST_TEMPLATE_3_DECL( LocalMap, Noncontig, LO, GO, NT )
{
  typedef Tpetra::Map<LO, GO, NT> map_type;
  int lclSuccess = 1;
  int gblSuccess = 1;

  Teuchos::OSTab tab0 (out);
  out << "LocalMap: Noncontiguous Map" << endl;
  Teuchos::OSTab tab1 (out);

  RCP<const Comm<int> > comm = getDefaultComm ();
  const int myRank = comm->getRank ();

  // Process p gets 5 indices.
  const LO numLclElts = static_cast<LO> (5);
  const GST numGblElts = Tpetra::Details::OrdinalTraits<GST>::invalid ();
  const GO indexBase = 0;

  Teuchos::Array<GO> eltList (numLclElts);
  const GO myGidStart = static_cast<GO> ((myRank + 1) * numLclElts  - 1);
  for (LO lid = 0; lid < numLclElts; ++lid) {
    eltList[lid] = static_cast<GO> (myGidStart - lid);
  }

  map_type gblMap (numGblElts, eltList (), indexBase, comm);
  auto lclMap = gblMap.getLocalMap ();

  try {
    // Sanity check on the global Map.
    TEST_EQUALITY( gblMap.getNodeNumElements (), static_cast<size_t> (numLclElts) );
    TEST_EQUALITY( gblMap.getMinLocalIndex (), static_cast<LO> (0) );
    TEST_EQUALITY( gblMap.getMaxLocalIndex (), static_cast<LO> (numLclElts - 1) );

    // Test constants.
    TEST_EQUALITY( gblMap.getNodeNumElements (), static_cast<size_t> (lclMap.getNodeNumElements ()) );
    TEST_EQUALITY( gblMap.getIndexBase (), lclMap.getIndexBase () );
    TEST_EQUALITY( gblMap.isContiguous (), lclMap.isContiguous () );
    TEST_EQUALITY( gblMap.getMinLocalIndex (), lclMap.getMinLocalIndex () );
    TEST_EQUALITY( gblMap.getMaxLocalIndex (), lclMap.getMaxLocalIndex () );
    TEST_EQUALITY( gblMap.getMinGlobalIndex (), lclMap.getMinGlobalIndex () );
    TEST_EQUALITY( gblMap.getMaxGlobalIndex (), lclMap.getMaxGlobalIndex () );

    // Test global-to-local and local-to-global index conversion.
    if (numLclElts > 0) {
      for (LO lid = 0; lid < numLclElts; ++lid) {
        const GO expectedGid = gblMap.getGlobalElement (lid);
        TEST_EQUALITY( lclMap.getGlobalElement (lid), expectedGid );
        TEST_EQUALITY( lclMap.getLocalElement (expectedGid), lid );
      }
    }
  } catch (...) {
    // None of the above methods should throw exceptions, but we catch
    // just in case.
    success = false;
  }

  // Make sure that all processes succeeded.
  lclSuccess = success ? 1 : 0;
  reduceAll<int, int> (*comm, REDUCE_MIN, lclSuccess, outArg (gblSuccess));
  TEST_EQUALITY_CONST( gblSuccess, 1 );

  if (gblSuccess != 1) {
    out << "The comparison test between Map and LocalMap "
      "failed on at least one process." << endl;
    // Make sure that the test fails, even if Process 0 was fine.
    success = false;
    return;
  }
}

//
// INSTANTIATIONS
//

#define UNIT_TEST_GROUP( LO, GO, NT ) \
  TEUCHOS_UNIT_TEST_TEMPLATE_3_INSTANT( LocalMap, UniformContig, LO, GO, NT ) \
  TEUCHOS_UNIT_TEST_TEMPLATE_3_INSTANT( LocalMap, NonuniformContig, LO, GO, NT ) \
  TEUCHOS_UNIT_TEST_TEMPLATE_3_INSTANT( LocalMap, Noncontig, LO, GO, NT )

  TPETRA_ETI_MANGLING_TYPEDEFS()

  TPETRA_INSTANTIATE_LGN(UNIT_TEST_GROUP)

} // namespace (anonymous)


